@@ -64,12 +64,19 @@ def alias(name):
     return "".join(x for x in name if x.isalnum() or x == "_").lower()
 
 
+def migrate_config(config):
+    # migrate old "auth" string into new "auth" json subtree
+    if "auth" in config:
+        if isinstance(config["auth"], str):
+            config["auth"] = dict(method=config["auth"])
+
+
 def deep_update(d, u):
     for k, v in six.iteritems(u):
         dv = d.get(k, {})
-        if not isinstance(dv, collections.Mapping):
+        if not isinstance(dv, collections.abc.Mapping):
             d[k] = v
-        elif isinstance(v, collections.Mapping):
+        elif isinstance(v, collections.abc.Mapping):
             d[k] = deep_update(dv, v)
         else:
             d[k] = v
@@ -104,7 +111,7 @@ def set_loglevel(app, loglevel_string):
     )
     loglevels = {"WARN": logging.WARN, "INFO": logging.INFO, "DEBUG": logging.DEBUG}
     logging.getLogger().setLevel(loglevels[loglevel_string])
-    logger.warn("Loglevel-Test: This is a warn-message!")
+    logger.warning("Loglevel-Test: This is a warn-message!")
     logger.info("Loglevel-Test: This is an info-message!")
     logger.debug("Loglevel-Test: This is an debug-message!")
 
@@ -296,3 +303,20 @@ def notify_upgrade(app, flash):
             "info",
         )
     return app.specter.version.current
+
+
+def is_ip_private(ip):
+    # https://en.wikipedia.org/wiki/Private_network
+    priv_lo = re.compile(r"^127\.\d{1,3}\.\d{1,3}\.\d{1,3}$")
+    priv_24 = re.compile(r"^10\.\d{1,3}\.\d{1,3}\.\d{1,3}$")
+    priv_20 = re.compile(r"^192\.168\.\d{1,3}.\d{1,3}$")
+    priv_16 = re.compile(r"^172.(1[6-9]|2[0-9]|3[0-1]).[0-9]{1,3}.[0-9]{1,3}$")
+
+    res = (
+        ip == "localhost"
+        or priv_lo.match(ip)
+        or priv_24.match(ip)
+        or priv_20.match(ip)
+        or priv_16.match(ip)
+    )
+    return res is not None
