@@ -1,5 +1,10 @@
-import ast, json, os, time, base64, random, requests
+import ast, csv, json, os, time, base64, random, requests, logging
+from io import StringIO
+from werkzeug.wrappers import Response
+from datetime import datetime
+from numbers import Number
 from ..util.tx import decoderawtransaction
+from ..util.price_providers import get_price_at
 
 from flask import (
     Flask,
@@ -11,7 +16,7 @@ from flask import (
     jsonify,
     flash,
 )
-from flask_login import login_required
+from flask_login import login_required, current_user
 from ..util.descriptor import AddChecksum, Descriptor
 from ..helpers import (
     bcur2base64,
@@ -20,6 +25,7 @@ from ..helpers import (
     is_testnet,
     parse_wallet_data_import,
 )
+from ..persistence import delete_file
 from ..key import Key
 from ..specter import Specter
 from ..specter_error import SpecterError
@@ -42,11 +48,12 @@ wallets_endpoint = Blueprint("wallets_endpoint", __name__)
 @wallets_endpoint.route("/wallets_overview/")
 @login_required
 def wallets_overview():
-    idx = int(request.args.get("idx", default=0))
+    app.specter.check_blockheight()
+    for wallet in app.specter.wallet_manager.wallets.values():
+        wallet.get_balance()
+        wallet.check_utxo()
     return render_template(
         "wallet/wallets_overview.jinja",
-        idx=idx,
-        history=True,
         specter=app.specter,
         rand=rand,
     )
@@ -79,14 +86,14 @@ def new_wallet(wallet_type):
     wallet_types = ["simple", "multisig", "import_wallet"]
     if wallet_type not in wallet_types:
         flash("Unknown wallet type requested", "error")
-        return redirect(url_for("new_wallet_type"))
+        return redirect(url_for("wallets_endpoint.new_wallet_type"))
 
     err = None
     if request.method == "POST":
         action = request.form["action"]
         if action == "importwallet":
-            wallet_data = json.loads(request.form["wallet_data"].replace("'", "h"))
             try:
+                wallet_data = json.loads(request.form["wallet_data"].replace("'", "h"))
                 (
                     wallet_name,
                     recv_descriptor,
@@ -94,7 +101,7 @@ def new_wallet(wallet_type):
                 ) = parse_wallet_data_import(wallet_data)
             except Exception:
                 flash("Unsupported wallet import format", "error")
-                return redirect(url_for("new_wallet_type"))
+                return redirect(url_for("wallets_endpoint.new_wallet_type"))
             # get min of the two
             # if the node is still syncing
             # and the first block with tx is not there yet
@@ -118,13 +125,13 @@ def new_wallet(wallet_type):
                 )
                 if descriptor is None:
                     flash("Invalid wallet descriptor.", "error")
-                    return redirect(url_for("new_wallet_type"))
+                    return redirect(url_for("wallets_endpoint.new_wallet_type"))
             except:
                 flash("Invalid wallet descriptor.", "error")
-                return redirect(url_for("new_wallet_type"))
+                return redirect(url_for("wallets_endpoint.new_wallet_type"))
             if wallet_name in app.specter.wallet_manager.wallets_names:
                 flash("Wallet with the same name already exists", "error")
-                return redirect(url_for("new_wallet_type"))
+                return redirect(url_for("wallets_endpoint.new_wallet_type"))
 
             sigs_total = descriptor.multisig_N
             sigs_required = descriptor.multisig_M
@@ -203,9 +210,13 @@ def new_wallet(wallet_type):
                     )
                     keys.append(unknown_cosigner)
                     cosigners.append(device)
-                wallet = app.specter.wallet_manager.create_wallet(
-                    wallet_name, sigs_required, address_type, keys, cosigners
-                )
+                try:
+                    wallet = app.specter.wallet_manager.create_wallet(
+                        wallet_name, sigs_required, address_type, keys, cosigners
+                    )
+                except Exception as e:
+                    flash("Failed to create wallet: %r" % e, "error")
+                    return redirect(url_for("wallets_endpoint.new_wallet_type"))
                 wallet.keypoolrefill(0, wallet.IMPORT_KEYPOOL, change=False)
                 wallet.keypoolrefill(0, wallet.IMPORT_KEYPOOL, change=True)
                 wallet.import_labels(wallet_data.get("labels", {}))
@@ -332,9 +343,24 @@ def new_wallet(wallet_type):
                 )
 
             # create a wallet here
-            wallet = app.specter.wallet_manager.create_wallet(
-                wallet_name, sigs_required, address_type, keys, cosigners
-            )
+            try:
+                wallet = app.specter.wallet_manager.create_wallet(
+                    wallet_name, sigs_required, address_type, keys, cosigners
+                )
+            except Exception:
+                err = "Failed to create wallet..."
+                return render_template(
+                    "wallet/new_wallet/new_wallet_keys.jinja",
+                    purposes=purposes,
+                    cosigners=cosigners,
+                    wallet_type=wallet_type,
+                    sigs_total=len(devices),
+                    sigs_required=max(len(devices) * 2 // 3, 1),
+                    error=err,
+                    specter=app.specter,
+                    rand=rand,
+                )
+
             app.logger.info("Created Wallet %s" % wallet_name)
             rescan_blockchain = "rescanblockchain" in request.form
             if rescan_blockchain:
@@ -344,8 +370,12 @@ def new_wallet(wallet_type):
                 if "utxo" in request.form.get("full_rescan_option"):
                     explorer = None
                     if "use_explorer" in request.form:
-                        explorer = app.specter.get_default_explorer()
-                    wallet.rescanutxo(explorer)
+                        explorer = request.form["explorer_url"]
+                    wallet.rescanutxo(
+                        explorer,
+                        app.specter.requests_session(explorer),
+                        app.specter.only_tor,
+                    )
                     app.specter.info["utxorescan"] = 1
                     app.specter.utxorescanwallet = wallet.alias
                 else:
@@ -402,77 +432,29 @@ def wallet(wallet_alias):
         app.logger.error("SpecterError while wallet: %s" % se)
         return render_template("base.jinja", error=se, specter=app.specter, rand=rand)
     if wallet.fullbalance > 0:
-        return redirect(
-            url_for("wallets_endpoint.tx_history", wallet_alias=wallet_alias)
-        )
+        return redirect(url_for("wallets_endpoint.history", wallet_alias=wallet_alias))
     else:
         return redirect(url_for("wallets_endpoint.receive", wallet_alias=wallet_alias))
 
 
 ###### Wallet transaction history ######
-
-
-@wallets_endpoint.route("/wallet/<wallet_alias>/tx/")
+@wallets_endpoint.route("/wallet/<wallet_alias>/history/")
 @login_required
-def tx(wallet_alias):
-    return redirect(url_for("wallets_endpoint.tx_history", wallet_alias=wallet_alias))
-
-
-@wallets_endpoint.route("/wallet/<wallet_alias>/tx/history/")
-@login_required
-def tx_history(wallet_alias):
+def history(wallet_alias):
     try:
         wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)
     except SpecterError as se:
         app.logger.error("SpecterError while wallet_tx: %s" % se)
         return render_template("base.jinja", error=se, specter=app.specter, rand=rand)
     # update balances in the wallet
+    app.specter.check_blockheight()
     wallet.get_balance()
-    idx = int(request.args.get("idx", default=0))
-
-    return render_template(
-        "wallet/history/txs/wallet_tx.jinja",
-        idx=idx,
-        wallet_alias=wallet_alias,
-        wallet=wallet,
-        history=True,
-        specter=app.specter,
-        rand=rand,
-    )
-
-
-@wallets_endpoint.route("/wallet/<wallet_alias>/tx/utxo/", methods=["GET", "POST"])
-@login_required
-def tx_utxo(wallet_alias):
-    try:
-        wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)
-    except SpecterError as se:
-        app.logger.error("SpecterError while wallet_addresses: %s" % se)
-        return render_template("base.jinja", error=se, specter=app.specter, rand=rand)
-    # update balances in the wallet
-    wallet.get_balance()
-    # check utxo list
     wallet.check_utxo()
-    viewtype = "address" if request.args.get("view") != "label" else "label"
-    idx = int(request.args.get("idx", default=0))
-    if request.method == "POST":
-        action = request.form["action"]
-        if action == "updatelabel":
-            label = request.form["label"]
-            account = request.form["account"]
-            if viewtype == "address":
-                wallet.setlabel(account, label)
-            else:
-                for address in wallet.addresses_on_label(account):
-                    wallet.setlabel(address, label)
-                wallet.getdata()
+
     return render_template(
-        "wallet/history/utxo/wallet_utxo.jinja",
-        idx=idx,
+        "wallet/history/wallet_history.jinja",
         wallet_alias=wallet_alias,
         wallet=wallet,
-        history=False,
-        viewtype=viewtype,
         specter=app.specter,
         rand=rand,
     )
@@ -499,20 +481,10 @@ def receive(wallet_alias):
     # check that current address is unused
     # and generate new one if it is
     wallet.check_unused()
-    history_idx = int(request.args.get("history_idx", default=0))
-    past_addresses = wallet.addresses[
-        -10 * history_idx - 2 : -10 * (history_idx + 1) - 2 : -1
-    ]
     return render_template(
         "wallet/receive/wallet_receive.jinja",
         wallet_alias=wallet_alias,
         wallet=wallet,
-        past_addresses=past_addresses,
-        past_descriptors=[
-            wallet.get_descriptor(address=addr) for addr in past_addresses
-        ],
-        addresses_count=len(wallet.addresses),
-        history_idx=history_idx,
         specter=app.specter,
         rand=rand,
     )
@@ -581,6 +553,12 @@ def send_new(wallet_alias):
                         amounts.append(float(output.split(",")[1].strip()) / 1e8)
                     else:
                         amounts.append(float(output.split(",")[1].strip()))
+            addresses = [
+                address.lower()
+                if address.startswith(("BC1", "TB1", "BCRT1"))
+                else address
+                for address in addresses
+            ]
             subtract = bool(request.form.get("subtract", False))
             subtract_from = int(request.form.get("subtract_from", 1)) - 1
             rbf = bool(request.form.get("rbf", False))
@@ -780,6 +758,36 @@ def import_psbt(wallet_alias):
     )
 
 
+###### Wallet addresses ######
+
+
+@wallets_endpoint.route("/wallet/<wallet_alias>/addresses/", methods=["GET"])
+@login_required
+def addresses(wallet_alias):
+    """Show informations about cached addresses (wallet._addresses) of the <wallet_alias>.
+    It updates balances in the wallet before renderization in order to show updated UTXO and
+    balance of each address."""
+
+    try:
+        wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)
+    except SpecterError as se:
+        app.logger.error("SpecterError while wallet_send: %s" % se)
+        return render_template("base.jinja", error=se, specter=app.specter, rand=rand)
+
+    # update balances in the wallet
+    app.specter.check_blockheight()
+    wallet.get_balance()
+    wallet.check_utxo()
+
+    return render_template(
+        "wallet/addresses/wallet_addresses.jinja",
+        wallet_alias=wallet_alias,
+        wallet=wallet,
+        specter=app.specter,
+        rand=rand,
+    )
+
+
 ###### Wallet settings ######
 
 
@@ -797,6 +805,8 @@ def settings(wallet_alias):
         if action == "rescanblockchain":
             startblock = int(request.form["startblock"])
             try:
+                delete_file(wallet._transactions.path)
+                wallet.fetch_transactions()
                 res = wallet.rpc.rescanblockchain(startblock, timeout=1)
             except requests.exceptions.ReadTimeout:
                 # this is normal behaviour in our usecase
@@ -813,8 +823,10 @@ def settings(wallet_alias):
         elif action == "rescanutxo":
             explorer = None
             if "use_explorer" in request.form:
-                explorer = app.specter.get_default_explorer()
-            wallet.rescanutxo(explorer)
+                explorer = request.form["explorer_url"]
+            wallet.rescanutxo(
+                explorer, app.specter.requests_session(explorer), app.specter.only_tor
+            )
             app.specter.info["utxorescan"] = 1
             app.specter.utxorescanwallet = wallet.alias
         elif action == "abortrescanutxo":
@@ -867,6 +879,7 @@ def settings(wallet_alias):
 
 
 ################## Wallet util endpoints #######################
+# TODO: move these to an API endpoint
 
 
 @wallets_endpoint.route("/wallet/<wallet_alias>/combine/", methods=["POST"])
@@ -885,6 +898,8 @@ def combine(wallet_alias):
     combined = None
 
     for i, psbt in enumerate(psbts):
+        if not psbt:
+            return "Cannot parse empty data as PSBT", 500
         if "UR:BYTES/" in psbt.upper():
             psbt = bcur2base64(psbt).decode()
 
@@ -960,7 +975,20 @@ def decoderawtx(wallet_alias):
         txid = request.form.get("txid", "")
         if txid:
             tx = wallet.rpc.gettransaction(txid)
-            return {"success": True, "tx": tx, "rawtx": decoderawtransaction(tx["hex"])}
+            # This is a fix for Bitcoin Core versions < v0.20
+            # These do not return the blockheight as part of the `gettransaction` command
+            # So here we check if this property is lacking and if so
+            # query the blockheader based on the transaction blockhash
+            ##################### Remove from here after dropping Core v0.19 support #####################
+            if "blockhash" in tx and "blockheight" not in tx:
+                tx["blockheight"] = wallet.rpc.getblockheader(tx["blockhash"])["height"]
+            ##################### Remove until here after dropping Core v0.19 support #####################
+            return {
+                "success": True,
+                "tx": tx,
+                "rawtx": decoderawtransaction(tx["hex"], app.specter.chain),
+                "walletName": wallet.name,
+            }
     except Exception as e:
         app.logger.warning("Failed to fetch transaction data. Exception: {}".format(e))
     return {"success": False}
@@ -981,3 +1009,672 @@ def rescan_progress(wallet_alias):
     except SpecterError as se:
         app.logger.error("SpecterError while get wallet rescan_progress: %s" % se)
         return {}
+
+
+@wallets_endpoint.route("/wallet/<wallet_alias>/get_label", methods=["POST"])
+@login_required
+def get_label(wallet_alias):
+    try:
+        wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)
+        address = request.form.get("address", "")
+        label = wallet.getlabel(address)
+        return {
+            "address": address,
+            "label": label,
+        }
+    except Exception as e:
+        logging.exception(e)
+        return "Error while get_label: %s" % e, 500
+
+
+@wallets_endpoint.route("/wallet/<wallet_alias>/set_label", methods=["POST"])
+@login_required
+def set_label(wallet_alias):
+    try:
+        wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)
+        address = request.form["address"]
+        label = request.form["label"]
+        wallet.setlabel(address, label)
+        return {"success": True}
+    except Exception as e:
+        logging.exception(e)
+        return "Error while set_label: %s" % e, 500
+
+
+@wallets_endpoint.route("/wallet/<wallet_alias>/txlist", methods=["POST"])
+@login_required
+def txlist(wallet_alias):
+    try:
+        wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)
+        idx = int(request.form.get("idx", 0))
+        limit = int(request.form.get("limit", 100))
+        search = request.form.get("search", None)
+        sortby = request.form.get("sortby", None)
+        sortdir = request.form.get("sortdir", "asc")
+        fetch_transactions = request.form.get("fetch_transactions", False)
+        txlist = wallet.txlist(
+            fetch_transactions=fetch_transactions,
+            validate_merkle_proofs=app.specter.config.get(
+                "validate_merkle_proofs", False
+            ),
+            current_blockheight=app.specter.info["blocks"],
+        )
+        return process_txlist(
+            txlist, idx=idx, limit=limit, search=search, sortby=sortby, sortdir=sortdir
+        )
+    except Exception as e:
+        logging.exception(e)
+        return "Error while getting txlist: %s" % e, 500
+
+
+@wallets_endpoint.route("/wallet/<wallet_alias>/utxo_list", methods=["POST"])
+@login_required
+def utxo_list(wallet_alias):
+    try:
+        wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)
+        idx = int(request.form.get("idx", 0))
+        limit = int(request.form.get("limit", 100))
+        search = request.form.get("search", None)
+        sortby = request.form.get("sortby", None)
+        sortdir = request.form.get("sortdir", "asc")
+        txlist = wallet.utxo
+        for tx in txlist:
+            if not tx.get("label", None):
+                tx["label"] = wallet.getlabel(tx["address"])
+        return process_txlist(
+            txlist, idx=idx, limit=limit, search=search, sortby=sortby, sortdir=sortdir
+        )
+    except Exception as e:
+        logging.exception(e)
+        return "Error while getting utxo list: %s" % e, 500
+
+
+@wallets_endpoint.route("/wallets_overview/txlist", methods=["POST"])
+@login_required
+def wallets_overview_txlist():
+    try:
+        idx = int(request.form.get("idx", 0))
+        limit = int(request.form.get("limit", 100))
+        search = request.form.get("search", None)
+        sortby = request.form.get("sortby", None)
+        sortdir = request.form.get("sortdir", "asc")
+        fetch_transactions = request.form.get("fetch_transactions", False)
+        txlist = app.specter.wallet_manager.full_txlist(
+            fetch_transactions=fetch_transactions,
+            validate_merkle_proofs=app.specter.config.get(
+                "validate_merkle_proofs", False
+            ),
+            current_blockheight=app.specter.info["blocks"],
+        )
+        return process_txlist(
+            txlist, idx=idx, limit=limit, search=search, sortby=sortby, sortdir=sortdir
+        )
+    except Exception as e:
+        logging.exception(e)
+        return "Error while getting full txlist: %s" % e, 500
+
+
+@wallets_endpoint.route("/wallets_overview/utxo_list", methods=["POST"])
+@login_required
+def wallets_overview_utxo_list():
+    try:
+        idx = int(request.form.get("idx", 0))
+        limit = int(request.form.get("limit", 100))
+        search = request.form.get("search", None)
+        sortby = request.form.get("sortby", None)
+        sortdir = request.form.get("sortdir", "asc")
+        fetch_transactions = request.form.get("fetch_transactions", False)
+        txlist = app.specter.wallet_manager.full_utxo()
+        return process_txlist(
+            txlist, idx=idx, limit=limit, search=search, sortby=sortby, sortdir=sortdir
+        )
+    except Exception as e:
+        logging.exception(e)
+        return "Error while getting full utxo list: %s" % e, 500
+
+
+@wallets_endpoint.route("/wallet/<wallet_alias>/addresses_list/", methods=["POST"])
+@login_required
+def addresses_list(wallet_alias):
+    """Return a JSON with keys:
+        addressesList: list of addresses with the properties
+                       (index, address, label, used, utxo, amount)
+        pageCount: total number of pages
+    POST parameters:
+        idx: pagination index (current page)
+        limit: maximum number of items on the page
+        sortby: field by which the list will be ordered
+                (index, address, label, used, utxo, amount)
+        sortdir: 'asc' (ascending) or 'desc' (descending) order
+        addressType: the current tab address type ('receive' or 'change')"""
+    try:
+        wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)
+
+        idx = int(request.form.get("idx", 0))
+        limit = int(request.form.get("limit", 100))
+        sortby = request.form.get("sortby", None)
+        sortdir = request.form.get("sortdir", "asc")
+        address_type = request.form.get("addressType", "receive")
+
+        addresses_list = wallet.addresses_info(address_type == "change")
+
+        result = process_addresses_list(
+            addresses_list, idx=idx, limit=limit, sortby=sortby, sortdir=sortdir
+        )
+
+        return {
+            "addressesList": json.dumps(result["addressesList"]),
+            "pageCount": result["pageCount"],
+        }
+
+    except Exception as e:
+        logging.exception(e)
+        return "Error while getting addresses_list: %s" % e, 500
+
+
+@wallets_endpoint.route("/wallet/<wallet_alias>/addressinfo/", methods=["POST"])
+@login_required
+def addressinfo(wallet_alias):
+    try:
+        wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)
+        address = request.form.get("address", "")
+        if address:
+            descriptor = wallet.get_descriptor(address=address)
+            address_info = wallet.get_address_info(address=address)
+            return {
+                "success": True,
+                "address": address,
+                "descriptor": descriptor,
+                "walletName": wallet.name,
+                **address_info,
+            }
+    except Exception as e:
+        app.logger.warning("Failed to fetch address data. Exception: {}".format(e))
+    return {"success": False}
+
+
+################## Wallet CSV export data endpoints #######################
+# Export wallet addresses list
+@wallets_endpoint.route("/wallet/<wallet_alias>/addresses_list.csv")
+@login_required
+def addresses_list_csv(wallet_alias):
+    """Return a CSV with addresses of the <wallet_alias> containing the
+    information: index, address, type, label, used, utxo and amount
+    of each of them.
+    GET parameters: sortby: field by which the CSV will be ordered
+                            (index, address, label, used, utxo, amount)
+                    sortdir: 'asc' (ascending) or 'desc' (descending) order
+                    address_type: the current tab address type ('receive' or 'change')
+                    onlyCurrentType: show all addresses (if false) or just the current
+                                     type (address_type param)"""
+
+    try:
+        wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)
+
+        sortby = request.args.get("sortby", "index")
+        sortdir = request.args.get("sortdir", "asc")
+        address_type = request.args.get("addressType", "receive")
+        only_current_type = request.args.get("onlyCurrentType", "false") == "true"
+
+        if not only_current_type:
+            receive_list = wallet.addresses_info(False)
+            change_list = wallet.addresses_info(True)
+
+            receive_result = process_addresses_list(
+                receive_list, idx=0, limit=0, sortby=sortby, sortdir=sortdir
+            )
+
+            change_result = process_addresses_list(
+                change_list, idx=0, limit=0, sortby=sortby, sortdir=sortdir
+            )
+
+            addressesList = (
+                receive_result["addressesList"] + change_result["addressesList"]
+            )
+        else:
+            addresses_list = wallet.addresses_info(address_type == "change")
+
+            result = process_addresses_list(
+                addresses_list, idx=0, limit=0, sortby=sortby, sortdir=sortdir
+            )
+
+            addressesList = result["addressesList"]
+
+        # stream the response as the data is generated
+        response = Response(
+            wallet_addresses_list_to_csv(addressesList),
+            mimetype="text/csv",
+        )
+        # add a filename
+        response.headers.set(
+            "Content-Disposition", "attachment", filename="addresses_list.csv"
+        )
+        return response
+    except Exception as e:
+        app.logger.error("Failed to export addresses list. Error: %s" % e)
+        flash("Failed to export addresses list. Error: %s" % e, "error")
+        return redirect(url_for("index"))
+
+
+# Export wallet transaction history
+@wallets_endpoint.route("/wallet/<wallet_alias>/transactions.csv")
+@login_required
+def tx_history_csv(wallet_alias):
+    try:
+        wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)
+        validate_merkle_proofs = app.specter.config.get("validate_merkle_proofs", False)
+        txlist = wallet.txlist(validate_merkle_proofs=validate_merkle_proofs)
+        search = request.args.get("search", None)
+        sortby = request.args.get("sortby", "time")
+        sortdir = request.args.get("sortdir", "desc")
+        txlist = json.loads(
+            process_txlist(
+                txlist, idx=0, limit=0, search=search, sortby=sortby, sortdir=sortdir
+            )["txlist"]
+        )
+        includePricesHistory = request.args.get("exportPrices", "false") == "true"
+
+        # stream the response as the data is generated
+        response = Response(
+            txlist_to_csv(
+                wallet, txlist, app.specter, current_user, includePricesHistory
+            ),
+            mimetype="text/csv",
+        )
+        # add a filename
+        response.headers.set(
+            "Content-Disposition", "attachment", filename="transactions.csv"
+        )
+        return response
+    except Exception as e:
+        logging.exception(e)
+        return "Failed to export wallet history. Error: %s" % e, 500
+
+
+# Export wallet UTXO list
+@wallets_endpoint.route("/wallet/<wallet_alias>/utxo.csv")
+@login_required
+def utxo_csv(wallet_alias):
+    try:
+        wallet = app.specter.wallet_manager.get_by_alias(wallet_alias)
+        includePricesHistory = request.args.get("exportPrices", "false") == "true"
+        search = request.args.get("search", None)
+        sortby = request.args.get("sortby", "time")
+        sortdir = request.args.get("sortdir", "desc")
+        txlist = json.loads(
+            process_txlist(
+                wallet.utxo,
+                idx=0,
+                limit=0,
+                search=search,
+                sortby=sortby,
+                sortdir=sortdir,
+            )["txlist"]
+        )
+        # stream the response as the data is generated
+        response = Response(
+            txlist_to_csv(
+                wallet,
+                txlist,
+                app.specter,
+                current_user,
+                includePricesHistory,
+            ),
+            mimetype="text/csv",
+        )
+        # add a filename
+        response.headers.set("Content-Disposition", "attachment", filename="utxo.csv")
+        return response
+    except Exception as e:
+        logging.exception(e)
+        return "Failed to export wallet utxo. Error: %s" % e, 500
+
+
+# Export all wallets transaction history combined
+@wallets_endpoint.route("/wallets_overview/full_transactions.csv")
+@login_required
+def wallet_overview_txs_csv():
+    try:
+        validate_merkle_proofs = app.specter.config.get("validate_merkle_proofs", False)
+        txlist = app.specter.wallet_manager.full_txlist(
+            validate_merkle_proofs=validate_merkle_proofs
+        )
+        search = request.args.get("search", None)
+        sortby = request.args.get("sortby", "time")
+        sortdir = request.args.get("sortdir", "desc")
+        txlist = json.loads(
+            process_txlist(
+                txlist, idx=0, limit=0, search=search, sortby=sortby, sortdir=sortdir
+            )["txlist"]
+        )
+        includePricesHistory = request.args.get("exportPrices", "false") == "true"
+        # stream the response as the data is generated
+        response = Response(
+            txlist_to_csv(
+                None, txlist, app.specter, current_user, includePricesHistory
+            ),
+            mimetype="text/csv",
+        )
+        # add a filename
+        response.headers.set(
+            "Content-Disposition", "attachment", filename="full_transactions.csv"
+        )
+        return response
+    except Exception as e:
+        logging.exception(e)
+        return "Failed to export wallets overview history. Error: %s" % e, 500
+
+
+# Export all wallets transaction history combined
+@wallets_endpoint.route("/wallets_overview/full_utxo.csv")
+@login_required
+def wallet_overview_utxo_csv():
+    try:
+        txlist = app.specter.wallet_manager.full_utxo()
+        search = request.args.get("search", None)
+        sortby = request.args.get("sortby", "time")
+        sortdir = request.args.get("sortdir", "desc")
+        txlist = json.loads(
+            process_txlist(
+                txlist, idx=0, limit=0, search=search, sortby=sortby, sortdir=sortdir
+            )["txlist"]
+        )
+        includePricesHistory = request.args.get("exportPrices", "false") == "true"
+        # stream the response as the data is generated
+        response = Response(
+            txlist_to_csv(
+                None, txlist, app.specter, current_user, includePricesHistory
+            ),
+            mimetype="text/csv",
+        )
+        # add a filename
+        response.headers.set(
+            "Content-Disposition", "attachment", filename="full_utxo.csv"
+        )
+        return response
+    except Exception as e:
+        logging.exception(e)
+        return "Failed to export wallets overview utxo. Error: %s" % e, 500
+
+
+################## Helpers #######################
+
+# Transactions list to user-friendly CSV format
+def txlist_to_csv(wallet, _txlist, specter, current_user, includePricesHistory=False):
+    txlist = []
+    for tx in _txlist:
+        if isinstance(tx["address"], list):
+            _tx = tx.copy()
+            for i in range(0, len(tx["address"])):
+                _tx["address"] = tx["address"][i]
+                _tx["amount"] = tx["amount"][i]
+                txlist.append(_tx.copy())
+        else:
+            txlist.append(tx.copy())
+    data = StringIO()
+    w = csv.writer(data)
+    # write header
+    symbol = "USD"
+    if specter.price_provider.endswith("_eur"):
+        symbol = "EUR"
+    elif specter.price_provider.endswith("_gbp"):
+        symbol = "GBP"
+    row = (
+        "Date",
+        "Label",
+        "Category",
+        "Amount ({})".format(specter.unit.upper()),
+        "Value ({})".format(symbol),
+        "Rate (BTC/{})".format(symbol)
+        if specter.unit != "sat"
+        else "Rate ({}/SAT)".format(symbol),
+        "TxID",
+        "Address",
+        "Block Height",
+        "Timestamp",
+        "Raw Transaction",
+    )
+    if not wallet:
+        row = ("Wallet",) + row
+    w.writerow(row)
+    yield data.getvalue()
+    data.seek(0)
+    data.truncate(0)
+
+    # write each log item
+    _wallet = wallet
+    for tx in txlist:
+        if not wallet:
+            wallet_alias = tx.get("wallet_alias", None)
+            try:
+                _wallet = specter.wallet_manager.get_by_alias(wallet_alias)
+            except Exception as e:
+                continue
+        label = _wallet.getlabel(tx["address"])
+        if label == tx["address"]:
+            label = ""
+        tx_raw = _wallet.gettransaction(tx["txid"])
+        if tx_raw:
+            tx_hex = tx_raw["hex"]
+        else:
+            tx_hex = ""
+        if not tx.get("blockheight", None):
+            if tx_raw.get("blockheight", None):
+                tx["blockheight"] = tx_raw["blockheight"]
+            else:
+                tx["blockheight"] = "Unconfirmed"
+        if specter.unit == "sat":
+            value = float(tx["amount"])
+            tx["amount"] = round(value * 1e8)
+        if includePricesHistory:
+            success, rate, symbol = get_price_at(
+                specter, current_user, timestamp=tx["time"]
+            )
+        else:
+            success = False
+        if success:
+            rate = float(rate)
+            if specter.unit == "sat":
+                rate = rate / 1e8
+            amount_price = float(tx["amount"]) * rate
+            if specter.unit == "sat":
+                rate = round(1 / rate)
+        else:
+            amount_price = None
+            rate = "-"
+
+        row = (
+            time.strftime("%Y-%m-%d", time.localtime(tx["time"])),
+            label,
+            tx["category"],
+            round(tx["amount"], (0 if specter.unit == "sat" else 8)),
+            round(amount_price * 100) / 100 if amount_price is not None else "-",
+            rate,
+            tx["txid"],
+            tx["address"],
+            tx["blockheight"],
+            tx["time"],
+            tx_hex,
+        )
+        if not wallet:
+            row = (tx.get("wallet_alias", ""),) + row
+        w.writerow(row)
+        yield data.getvalue()
+        data.seek(0)
+        data.truncate(0)
+
+
+# Addresses list to user-friendly CSV format
+def addresses_list_to_csv(wallet):
+    data = StringIO()
+    w = csv.writer(data)
+    # write header
+    row = (
+        "Address",
+        "Label",
+        "Index",
+        "Used",
+        "Current balance",
+    )
+    w.writerow(row)
+    yield data.getvalue()
+    data.seek(0)
+    data.truncate(0)
+
+    # write each log item
+    for address in wallet._addresses:
+        address_info = wallet.get_address_info(address)
+        if not address_info.is_labeled and not address_info.used:
+            continue
+        row = (
+            address,
+            address_info.label,
+            "(external)"
+            if address_info.is_external
+            else (
+                str(address_info.index) + (" (change)" if address_info.change else "")
+            ),
+            address_info.used,
+        )
+        if address_info.is_external:
+            balance_on_address = "unknown (external address)"
+        else:
+            balance_on_address = 0
+            if address_info.used:
+                for tx in wallet.utxo:
+                    if tx.get("address", "") == address:
+                        balance_on_address += tx.get("amount", 0)
+        row += (balance_on_address,)
+
+        w.writerow(row)
+        yield data.getvalue()
+        data.seek(0)
+        data.truncate(0)
+
+
+def wallet_addresses_list_to_csv(addresses_list):
+    """Convert a list of the wallet addresses to user-friendly CSV format
+    Parameters: addresses_list: a dict of addresses informations
+                (index, address, type, label, used, utxo and amount)"""
+    data = StringIO()
+    w = csv.writer(data)
+    # write header
+    row = (
+        "Index",
+        "Address",
+        "Type",
+        "Label",
+        "Used",
+        "UTXO",
+        "Amount (BTC)",
+    )
+    w.writerow(row)
+    yield data.getvalue()
+    data.seek(0)
+    data.truncate(0)
+
+    # write each log item
+    for address_item in addresses_list:
+
+        used = "Yes" if address_item["used"] else "No"
+
+        row = (
+            address_item["index"],
+            address_item["address"],
+            address_item["type"],
+            address_item["label"],
+            used,
+            address_item["utxo"],
+            address_item["amount"],
+        )
+
+        w.writerow(row)
+        yield data.getvalue()
+        data.seek(0)
+        data.truncate(0)
+
+
+def process_txlist(txlist, idx=0, limit=100, search=None, sortby=None, sortdir="asc"):
+    if search:
+        txlist = [
+            tx
+            for tx in txlist
+            if search in tx["txid"]
+            or (
+                any(search in address for address in tx["address"])
+                if isinstance(tx["address"], list)
+                else search in tx["address"]
+            )
+            or (
+                any(search in label for label in tx["label"])
+                if isinstance(tx["label"], list)
+                else search in tx["label"]
+            )
+            or (
+                any(search in str(amount) for amount in tx["amount"])
+                if isinstance(tx["amount"], list)
+                else search in str(tx["amount"])
+            )
+            or search in str(tx["confirmations"])
+            or search in str(tx["time"])
+            or search
+            in str(format(datetime.fromtimestamp(tx["time"]), "%d.%m.%Y %H:%M"))
+        ]
+    if sortby:
+
+        def sort(tx):
+            val = tx.get(sortby, None)
+            final = val
+            if val:
+                if isinstance(val, list):
+                    if isinstance(val[0], Number):
+                        final = sum(val)
+                    elif isinstance(val[0], str):
+                        final = sorted(
+                            val, key=lambda s: s.lower(), reverse=sortdir != "asc"
+                        )[0].lower()
+                elif isinstance(val, str):
+                    final = val.lower()
+            return final
+
+        txlist = sorted(txlist, key=sort, reverse=sortdir != "asc")
+    if limit:
+        page_count = (len(txlist) // limit) + (0 if len(txlist) % limit == 0 else 1)
+        txlist = txlist[limit * idx : limit * (idx + 1)]
+    else:
+        page_count = 1
+    return {"txlist": json.dumps(txlist), "pageCount": page_count}
+
+
+def process_addresses_list(
+    addresses_list, idx=0, limit=100, sortby=None, sortdir="asc"
+):
+    """Receive an address list as parameter and sort it or slice it for pagination.
+    Parameters: addresses_list: list of dict with the keys
+                                (index, address, label, used, utxo, amount)
+                idx: pagination index (current page)
+                limit: maximum number of items on the page
+                sortby: field by which the list will be ordered
+                        (index, address, label, used, utxo, amount)
+                sortdir: 'asc' (ascending) or 'desc' (descending) order"""
+    if sortby:
+
+        def sort(addr):
+            val = addr.get(sortby, None)
+            final = val
+            if val:
+                if isinstance(val, str):
+                    final = val.lower()
+            return final
+
+        addresses_list = sorted(addresses_list, key=sort, reverse=sortdir != "asc")
+
+    if limit:
+        page_count = (len(addresses_list) // limit) + (
+            0 if len(addresses_list) % limit == 0 else 1
+        )
+        addresses_list = addresses_list[limit * idx : limit * (idx + 1)]
+    else:
+        page_count = 1
+
+    return {"addressesList": addresses_list, "pageCount": page_count}
